<?php

namespace WebServiceCall\Connection;

use WebServiceCall\Log\FileLogger;
use WebServiceCall\Log\OutputStreamLogger;

class Connection implements IOracleConnHandler {
    private const AUTO_COMMIT_ROWS_LIMIT = 500;
    private const NRORG_NOT_FOUND = "Não foi possível obter o NRORG !";
    
    private $conn = null;

    private static $instance;
    private static $rowBackCount;
    private $nrorg;
    
    private function __construct() {
        $this->conn = OCI8Connection::getConnection();
        self::$rowBackCount = 0;
    }

    private function __clone() {}
    
    private function __wakeup() {}
    
    public static function getInstance() {
        if(self::$instance === null){
            self::$instance = new self;
        }
        return self::$instance;
    }

    public function commit() {
        oci_commit($this->conn);
    }
    
    public function rollback() {
        oci_rollback($this->conn);
    }
    
    public function closeConn() {
        oci_close($this->conn);
    }
    
    private function getMaxLength($dataType, $data) {
        if($dataType === SQLT_CHR) {
            return strlen($data);
        } else if($dataType === OCI_B_INT || $dataType === SQLT_INT ) {
            return 10;
        }
    }
    
    private function setSQLParams($stmt, $params, $types = []) {
        if(empty($types)) {
            foreach($params as $key => $value) {
                $bool = oci_bind_by_name($stmt, $key, $value);
                if(!$bool) {
                    $e = oci_error($stmt);
                    throw new \Exception("O parâmetro" . $key . " é inválido. " . $e['message']);
                }
            }
        } else {
            foreach($params as $key => $value) {
                $bool = oci_bind_by_name($stmt, $key, $value, $this->getMaxLength($types[$key], $value), $types[$key]);
                if(!$bool) {
                    $e = oci_error($stmt);
                    throw new \Exception("O parâmetro" . $key . " é inválido. " . $e['message']);
                }
            }
        }

        return $stmt;
    }
    
    private function buildStatement($sql, $params, $force = false) {
        $stmt = oci_parse($this->conn, $sql);
        oci_internal_debug(true);
        if (!$stmt) {
            $e = oci_error($this->conn);  // For oci_parse errors pass the connection handle
            throw new \Exception($e['message']);
        }
        if(!$force) {        
            $stmt = $this->setSQLParams($stmt, $params);
        }

        FileLogger::log($sql);
        
        return $stmt;
    }
    
    private function runSQL($sql, $params, $force = false) {
        $stmt = $this->buildStatement($sql, $params, $force);
        $runResult = oci_execute($stmt, OCI_NO_AUTO_COMMIT);
        
        if(!$runResult) {
            $e = oci_error();
            throw new \Exception("Não foi possível executar o comando SQL fornecido: " . $sql . PHP_EOL . $e['message']);
        }
        
        return $stmt;
    }
    
    private function afterRunSQL($stmt) {
        $stmt = $this->handleOCI8Error($stmt);
        oci_free_statement($stmt);
        $this->handleAutoCommit();
        return $stmt;
    }
    
    public function fetchAssoc($sql, array $params = [], $types = [], $force = false) {
        $result = NULL;
        try {
            if($force) {
                // When you don't want to worry with oci_bindings problems. 
                // I think that some oci extensions may be corrupted in our envs.
                // So I'm using this method instead of binding.
                $sql = $this->replaceParamsOnQuery($sql, $params, $types);
            }
            
            $stmt = oci_parse($this->conn, $sql);
            
            if(!$force) {        
                $stmt = $this->setSQLParams($stmt, $params, $types);
            }
            
            $stmt = $this->runSQL($sql, $params, $force);
            
            $result =  oci_fetch_assoc($stmt);
        } catch(\Exception $e) {
            OutputStreamLogger::log($e->getMessage() . $sql);
        } finally {
            if(!$stmt || !$result) {
                $e = oci_error();
                throw new \Exception("Não foi possível executar o comando SQL fornecido: " . $sql . PHP_EOL . $e['message']);
            }
            $stmt = $this->afterRunSQL($stmt);
        }
    
        return !empty($result) ? $result : null;
    }
    
    public function fetchRow($sql, array $params = [], $types = [], $force = false) {
        $result = NULL;
        try {
            if($force) {
                // When you don't want to worry with oci_bindings problems. 
                // I think that some oci extensions may be corrupted in our envs.
                // So I'm using this method instead of binding.
                $sql = $this->replaceParamsOnQuery($sql, $params, $types);
            }
            
            $stmt = oci_parse($this->conn, $sql);
            
            if(!$force) {        
                $stmt = $this->setSQLParams($stmt, $params, $types);
            }
            
            $stmt = $this->runSQL($sql, $params, $force);
            $result = oci_fetch_row($stmt);
        } catch(\Exception $e) {
            OutputStreamLogger::log($e->getMessage() . $sql);
        } finally {
            if(!$stmt) {
                throw new \Exception("Não foi possível executar o comando SQL fornecido: " . $sql . PHP_EOL);
            }
            $stmt = $this->afterRunSQL($stmt);
        }

        return !empty($result) ? $result[0] : null;
    }
    
    public function fetchAll($sql, array $params = [], $types = [], $force = false) {
        $result = NULL;
        try {
            if($force) {
                // When you don't want to worry with oci_bindings problems. 
                // I think that some oci extensions may be corrupted in our envs.
                // So I'm using this method instead of binding.
                $sql = $this->replaceParamsOnQuery($sql, $params, $types);
            }
            
            $stmt = oci_parse($this->conn, $sql);
            
            if(!$force) {        
                $stmt = $this->setSQLParams($stmt, $params, $types);
            }
            
            $stmt = $this->runSQL($sql, $params, $force);
            $nroOfRows = oci_fetch_all($stmt, $result);
            
        } catch(\Exception $e) {
            OutputStreamLogger::log($e->getMessage() . $sql);
        } finally {
            if(!$stmt) {
                throw new \Exception("Não foi possível executar o comando SQL fornecido: " . $sql . PHP_EOL);
            }
            $stmt = $this->afterRunSQL($stmt);
        }

        return !empty($result) ? $result : null;
    }
    
    private function replaceParamsOnQuery($sql, $params, $types = []) {
        foreach($params as $key => $value) {
            $sql = str_replace($key, $value, $sql);
        }
        
        return $sql;
    }
    
    // Use this for INSERT's and UPDATE's
    public function executeQuery($sql, array $params = [], $types = [], $force = false) {
        try {
            if($force) {
                // When you don't want to worry with oci_bindings problems. 
                // I think that some oci extensions may be corrupted in our envs.
                // So I'm using this method instead of binding.
                $sql = $this->replaceParamsOnQuery($sql, $params, $types);
            }
            
            $stmt = oci_parse($this->conn, $sql);
            
            if(!$force) {        
                $stmt = $this->setSQLParams($stmt, $params, $types);
            }
            
            $bool = oci_execute($stmt, OCI_NO_AUTO_COMMIT);
            
            if(!$bool) {
                $e = oci_error($stmt);
                OutputStreamLogger::log($e['message'], $sql);
            }
            
            self::$rowBackCount += oci_num_rows($stmt);
        } catch(\Exception $e) {
            OutputStreamLogger::log($e->getMessage() . $sql);
        } finally {
            $stmt = $this->afterRunSQL($stmt);
        }
        
        return $bool;
    }
    
    public function getNRORG() {
        if(!$this->nrorg) {
            $this->setNRORG();
        }
        
        return $this->nrorg;
    }
    
    private function setNRORG() {
        $result =  $this->fetchRow(self::GET_NRORG);
        
        if(empty($result)) {
            throw new \Exception(self::NRORG_NOT_FOUND);
        }
        $this->nrorg = $result;
    }
    
    /*
     * Gera um NRSEQUENCIAL para a tabela especifícada para ser usado em uma operação de insert.
     *
     * @param tableName: nome da tabela
     * @param nrSeq: NRSEQUENCIAL inicial. O default é não passar nenhum paramêtro para que se gere o NRSEQUENCIAL imediatamente
     * anterior ao mais recente na tabela NOVO_CODIGO
     * @param nrSeqLength: tamanho do inteiro gerado. O default é 20
     * @param qtd: quantidade de códigos que serão gerados. O default é sempre 1
     * @param nrorg: número da organização. O default é null
     *
     * @return o NRSEQUENCIAL gerado pela procedure que é armazenado na variável nrSeq
     */
    public function getNextTableId($tableName, $nrSeq = '', $nrSeqLength = 20, $qtd = 1) {
        $conn = $this->conn;
        $sql = 'CALL NOVO_CODIGO_NT(:P_CONTADOR, :P_SEQUENCIAL, :P_QTDE)';
        $statement = oci_parse($conn, $sql);
        oci_bind_by_name($statement, ':P_CONTADOR', $tableName);
        oci_bind_by_name($statement, ':P_SEQUENCIAL', $nrSeq, $nrSeqLength);
        oci_bind_by_name($statement, ':P_QTDE', $qtd);
        oci_execute($statement, OCI_NO_AUTO_COMMIT);
        
        oci_free_statement($statement);
        $this->handleOCI8Error($statement);
        
        if (!$nrSeq) {
            throw new \Exception("Não foi possível gerar chave primária para a tabela " . $tableName);
        }

        return $nrSeq;
    }
    
    private function handleOCI8Error($stmt) {
        $e = oci_error($stmt);
        if ($e) {
            throw new \Exception($e["message"]);
        }
        return $stmt;
    }
    
    public function getConn() {
        return $this->conn;
    }
    
    // Função que faz o commit sempre que o rowBack chega no limite. 
    public function handleAutoCommit() {
        if(self::$rowBackCount >= self::AUTO_COMMIT_ROWS_LIMIT) {
            $this->commit();
            self::$rowBackCount = 0;
        }
    }
    
    private const GET_NRORG = "
        SELECT NVL(DECODE(SYS_CONTEXT('nrorg_ctx', 'nrorg'), 0, 1, SYS_CONTEXT('nrorg_ctx', 'nrorg')), 1) NRORG 
          FROM DUAL
    ";
}